from .mixins import SuperSingleObjectMixin, PreProcessMixin
from django.core.exceptions import ImproperlyConfigured
from django.http import HttpResponse
from django.views.generic.base import View, RedirectView
from django.views.generic.list import ListView
import StringIO


__all__ = ['DownloadReportView', 'PaginatedListView']


class DownloadReportView(View):
    builder_class = None

    def get_builder(self):
        builder_class = self.get_builder_class()
        if builder_class is None:
            raise ImproperlyConfigured(u'You must define `builder_class` '
                                       'in your view')
        return builder_class(**self.get_builder_kwargs())

    def get_builder_class(self):
        return self.builder_class

    def get_builder_kwargs(self):
        return {}

    def get(self, *args, **kwargs):
        self.report = self.get_builder()
        output = StringIO.StringIO()
        self.report.generate(output)

        response = HttpResponse(content=output.getvalue(),
                                mimetype=self.get_mimetype())
        response['Content-Disposition'] = 'attachment;filename="%s"' % \
                                                            self.get_filename()
        return response

    def get_filename(self):
        return '%s.%s' % ('report', self.get_file_ext())

    def get_file_ext(self):
        return self.report.get_file_ext()

    def get_mimetype(self):
        return self.report.get_mimetype()


class ExportView(DownloadReportView):
    queryset = None

    def get_queryset(self):
        return self.queryset

    def get_builder_kwags(self):
        return {
            'queryset': self.get_queryset()
        }

    def post(self, *args, **kwargs):
        return self.get(*args, **kwargs)


class PaginatedListView(ListView):

    def get_context_data(self, **kwargs):
        ctx = ListView.get_context_data(self, **kwargs)
        ctx.update({
            'pages': self.get_pages(ctx['paginator'], ctx['page_obj'])
        })
        return ctx

    def get_pages(self, p, page):
        pags = []
        pg = page.number
        if 1 < pg - 2:
            pags.append((1, 1,))
            pags.append(('...', None,))
        #        if pg - 1 in p.page_range:
        #            pags.append(('Prev', p.page(pg - 1) ,))
        for n in range(pg - 2, pg + 2):
            if n in p.page_range:
            #                pags.append((n, p.page(n) ,))
                pags.append((n, n,))
        if p.num_pages not in range(pg - 2, pg + 3):
            pags.append(('...', None,))
            pags.append((p.num_pages, p.num_pages,))
        return pags


class ProcessView(PreProcessMixin, RedirectView):

    def pre_process(self):
        return self.process()

    def process(self):
        return


class ProcessObjectView(SuperSingleObjectMixin, RedirectView):

    def get_redirect_url(self, **kwargs):
        url = super(ProcessObjectView, self).get_redirect_url(**kwargs)
        if url is None:
            try:
                return self.object.get_absolute_url()
            except:
                pass
        return url

    def pre_process(self):
        SuperSingleObjectMixin.pre_process(self)
        return self.process()

    def process(self):
        pass



from .utils import LazyEncoder
from django.http.response import HttpResponse
import json


class JsonResponse(HttpResponse):
    def __init__(self, data):
        HttpResponse.__init__(
            self, content=json.dumps(data, cls=LazyEncoder),
            content_type="application/json",
        )



# major, minor, patch
VERSION = (0, 1, 2)


def get_version():
    version = '%s.%s' % (VERSION[0], VERSION[1])
    if VERSION[2]:
        version = '%s.%s' % (version, VERSION[2])
    return version



'''
Created on 12.11.2011

@author: alekam
'''
from django.contrib import messages
from django.core.exceptions import ImproperlyConfigured
from django.core.urlresolvers import reverse
from django.core.validators import EMPTY_VALUES
from django.db.models.query_utils import Q
from django.http import HttpResponseRedirect
from django.http.response import HttpResponseBase
from django.views.generic.detail import SingleObjectMixin


__all__ = ['SuperSingleObjectMixin', 'ShowSuccessMessageMixin']


class PreProcessMixin(object):
    """
    Mixin to do some actions before start processing GET, PUT or POST request
    """

    def dispatch(self, request, *args, **kwargs):
        # Try to dispatch to the right method; if a method doesn't exist,
        # defer to the error handler. Also defer to the error handler if the
        # request method isn't on the approved list.
        if request.method.lower() in self.http_method_names:
            handler = getattr(self, request.method.lower(),
                              self.http_method_not_allowed)
        else:
            handler = self.http_method_not_allowed
        self.request = request
        self.args = args
        self.kwargs = kwargs
        response = self.pre_process()
        if isinstance(response, HttpResponseBase):
            return response
        return handler(request, *args, **kwargs)

    def pre_process(self):
        """Can be overridden in child class"""
        return


class SuperSingleObjectMixin(PreProcessMixin, SingleObjectMixin):
    """
    This mixin is the same as SingleObjectMixin, but it set up
    `object` property before start processing request in GET or POST methods
    """
    def pre_process(self):
        self.object = self.get_object()


class ShowSuccessMessageMixin(object):
    success_message = None

    def show_success_message(self):
        message = self.get_success_message()
        if message is not None:
            messages.success(self.request, message)

    def get_success_message(self):
        if self.success_message is not None:
            return self.success_message
        return None


class GetQuerysetMixin(object):
    queryset = None
    model = None

    def get_queryset(self):
        """
        Get the list of items for this view. This must be an interable, and may
        be a queryset (in which qs-specific behavior will be enabled).
        """
        if self.queryset is not None:
            queryset = self.queryset
            if hasattr(queryset, '_clone'):
                queryset = queryset._clone()
        elif self.model is not None:
            queryset = self.model._default_manager.all()
        else:
            raise ImproperlyConfigured(u"'%s' must define 'queryset' or 'model'"
                                       % self.__class__.__name__)
        return queryset


class NextRedirectMixin(object):
    next_url_field_name = 'next'

    def get_success_url(self):
        return self.request.REQUEST.get(self.next_url_field_name)

    def is_safe_url(self, url):
        return is_safe_url(url=url, host=self.request.get_host())


class MassActionMixin(object):
    actions = {}

    def post(self, request, *args, **kwargs):
        action = self.request.POST.get('action', False)
        if action:
            ids = self.request.POST.getlist('id')
            qs = self.get_queryset().filter(id__in=ids)
            action = self.actions.get(action, False)
            if action:
                action(self, qs)
            if not request.is_ajax():
                return HttpResponseRedirect(reverse('report_list'))
        return self.get(request, *args, **kwargs)


class FilteredListMixin(object):
    filter_class = None

    def get_filter_class(self):
        return self.filter_class

    def get_filter_kwargs(self, qs=None):
        if qs is None:
            qs = self.get_queryset()
        return {
            'data': self.request.REQUEST,
            'queryset': qs
        }

    def get_filter_object(self, qs=None):
        return self.get_filter_class()(**self.get_filter_kwargs(qs))

    def get_context_data(self, **kwargs):
        qs = kwargs.get('object_list')
        if qs is None:
            qs = self.get_queryset()
        f = self.get_filter_object(qs)
        kwargs['filter'] = f
        kwargs['object_list'] = f.qs
        return kwargs


class SearchMixin(object):
    search_field = 'search'
    search_fields = []

    def get_queryset(self, qs):
        query = self.request.GET.get(self.search_field)
        if query not in EMPTY_VALUES:
            query = query.strip()
            if query not in EMPTY_VALUES and len(self.search_fields) > 0:
                lookup = None
                for field in self.search_fields:
                    if lookup is None:
                        lookup = Q(**{'%s__icontains' % field: query})
                    else:
                        lookup = lookup | Q(**{'%s__icontains' % field: query})
                qs = qs.filter(lookup)
        return qs

    def get_context_data(self, **kwargs):
        kwargs[self.search_field] = self.request.GET.get(self.search_field)
        return kwargs



'''
Created on 05.03.2012

@author: alekam
'''
from django.core.exceptions import ObjectDoesNotExist
from django.http import Http404
from django.shortcuts import get_object_or_404
from django.utils.encoding import smart_str
from django.views.generic.detail import DetailView
from django.views.generic.edit import CreateView, FormMixin, UpdateView
from django.views.generic.list import ListView


__all__ = ['InlineListView', 'CreateView', 'InlineUpdateView']


class InlineMixin(object):
    master_model = None
    context_master_object_name = None
    master_pk_url_kwarg = 'pk'

    def get_master_object(self, pk=None):
        if pk is None:
            pk = self.kwargs.get(self.master_pk_url_kwarg)
        return get_object_or_404(self.master_model, pk=pk)

    def get_context_master_object_name(self):
        if self.context_master_object_name:
            return self.context_master_object_name
        elif hasattr(self.master_object, 'model'):
            return smart_str(self.master_object.model._meta.object_name.lower())
        elif self.master_model is not None:
            return smart_str(self.master_model._meta.object_name.lower())
        else:
            return None

    def get_context_data(self, **kwargs):
        kwargs.update({
            'master': self.master_object,
            self.get_context_master_object_name(): self.master_object
        })
        return kwargs


class InlineListView(InlineMixin, ListView):

    def get(self, *args, **kwargs):
        self.master_object = self.get_master_object()
        return super(InlineListView, self).get(*args, **kwargs)

    def get_context_data(self, **kwargs):
        kwargs = ListView.get_context_data(self, **kwargs)
        kwargs = InlineMixin.get_context_data(self, **kwargs)
        return kwargs


class InlineFormMixin(InlineMixin):
    success_redirect_on_master = True
    master_field_name = None  # required field

    def form_valid(self, form):
        self.object = form.save(commit=False)
        self.save_object()
        return FormMixin.form_valid(self, form)

    def save_object(self):
        setattr(self.object, self.master_field_name, self.get_master_object())
        self.object.save()

    def get_master_object(self):
        try:
            return InlineMixin.get_master_object(self)
        except ObjectDoesNotExist:
            return None

    def get_success_url(self):
        if self.success_redirect_on_master:
            return self.get_master_object().get_absolute_url()
        return self.object.get_absolute_url()


class InlineCreateView(InlineFormMixin, CreateView):

    def get(self, request, *args, **kwargs):
        self.master_object = self.get_master_object()
        return CreateView.get(self, request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        self.master_object = self.get_master_object()
        return CreateView.post(self, request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        kwargs = CreateView.get_context_data(self, **kwargs)
        kwargs = InlineFormMixin.get_context_data(self, **kwargs)
        return kwargs


class InlineUpdateView(InlineFormMixin, UpdateView):

    def get(self, request, *args, **kwargs):
        self.master_object = self.get_master_object()
        return UpdateView.get(self, request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        self.master_object = self.get_master_object()
        return UpdateView.post(self, request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        kwargs = UpdateView.get_context_data(self, **kwargs)
        kwargs = InlineFormMixin.get_context_data(self, **kwargs)
        return kwargs

    def get_object(self, queryset=None):
        if queryset is None:
            queryset = self.get_queryset()
        queryset = queryset.filter(**self.get_lookup())
        return UpdateView.get_object(self, queryset)

    def get_lookup(self):
        return {
            '%s__id' % self.master_field_name: \
                                self.kwargs.get(self.master_pk_url_kwarg)
        }


class InlineDetailView(InlineMixin, DetailView):

    def get(self, *args, **kwargs):
        self.master_object = self.get_master_object()
        return super(InlineDetailView, self).get(*args, **kwargs)



from django.views.generic.edit import FormView


__all__ = ['FormWithRequestView', ]


class FormWithRequestView(FormView):

    def get_form_kwargs(self):
        kwargs = FormView.get_form_kwargs(self)
        kwargs['request'] = self.request
        return kwargs



'''
Generic views for formsets processing

Created on 08.01.2012

@author: alekam <alex@kamedov.ru>
'''
from .mixins import GetQuerysetMixin, SuperSingleObjectMixin
from django.core.exceptions import ImproperlyConfigured
from django.forms.formsets import BaseFormSet, formset_factory
from django.forms.models import ModelForm, BaseModelFormSet, \
    modelformset_factory, inlineformset_factory, BaseInlineFormSet
from django.utils.encoding import smart_str
from django.views.generic.base import View, TemplateResponseMixin
from django.views.generic.edit import FormMixin


class FormsetMixin(FormMixin):
    """
    A mixin that provides a way to show and handle a formset in a request.
    """
    formset_class = BaseFormSet
    formset_extra = 1
    can_order = False
    can_delete = False
    max_num = None
    initial = {}

    def get_formset_class(self):
        """
        Returns an instance of the formset to be used in this view.
        """
        kwargs = self.get_formset_class_kwargs()
        return formset_factory(**kwargs)

    def get_formset_class_kwargs(self):
        """
        Returns keywords arguments to create the formset class
        """
        return {
            'form': self.get_form_class(),
            'formset': self.formset_class,
            'extra': self.formset_extra,
            'can_order': self.can_order,
            'can_delete': self.can_delete,
            'max_num': self.max_num
        }

    def get_formset_kwargs(self):
        return FormMixin.get_form_kwargs(self)

    def get_formset(self, formset_class=None):
        if formset_class is None:
            formset_class = self.get_formset_class()
        return formset_class(**self.get_formset_kwargs())


class MultipleObjectWithoutPaginationMixin(GetQuerysetMixin):
    context_object_name = None

    def get_context_object_name(self, object_list):
        """
        Get the name of the item to be used in the context.
        """
        if self.context_object_name:
            return self.context_object_name
        elif hasattr(object_list, 'model'):
            return smart_str('%s_list' % object_list.model._meta.object_name.lower())
        else:
            return None

    def get_context_data(self, **kwargs):
        """
        Get the context for this view.
        """
        queryset = kwargs.pop('object_list')
        context_object_name = self.get_context_object_name(queryset)
        context = {
            'object_list': queryset
        }
        context.update(kwargs)
        if context_object_name is not None:
            context[context_object_name] = queryset
        return context


class BaseModelFormsetMixin(FormsetMixin):
    """
    Helpful class to process model formsets
    """
    model = None
    form_class = ModelForm
    formset_class = BaseModelFormSet
    formfield_callback = None
    fields = None
    exclude = None
    use_queryset = False

    def get_formset_class_kwargs(self):
        kwargs = FormsetMixin.get_formset_class_kwargs(self)
        kwargs.update({
            'model': self.model,
            'formfield_callback': self.formfield_callback,
            'fields': self.fields,
            'exclude': self.get_excluded_fields()
        })
        return kwargs

    def get_excluded_fields(self):
        return self.exclude

    def form_valid(self, formset):
        # TODO: this not work sometimes
        formset.save()
        return FormMixin.form_valid(self, formset)

    def form_invalid(self, formset):
        return self.render_to_response(self.get_context_data(formset=formset))


class ModelFormsetMixin(BaseModelFormsetMixin, GetQuerysetMixin):
    use_queryset = True

    def get_formset_class(self):
        """
        Returns an instance of the formset to be used in this view.
        """
        kwargs = self.get_formset_class_kwargs()

        model = kwargs['model']
        del kwargs['model']

        return modelformset_factory(model, **kwargs)

    def get_formset_kwargs(self):
        kwargs = FormsetMixin.get_formset_kwargs(self)
        if self.use_queryset:
            kwargs['queryset'] = self.get_queryset()
        return kwargs


class InlineModelFormsetMixin(BaseModelFormsetMixin, SuperSingleObjectMixin):
    # parent model
    model = None
    # model
    inline_model = None
    fk_name = None
    form_class = ModelForm
    formset_class = BaseInlineFormSet

    def get_formset_class_kwargs(self):
        kwargs = BaseModelFormsetMixin.get_formset_class_kwargs(self)
        kwargs.update({
            'parent_model': self.model,
            'model': self.inline_model,
            'fk_name': self.fk_name,
        })
        return kwargs

    def get_formset_class(self):
        """
        Returns an instance of the formset to be used in this view.
        """
        kwargs = self.get_formset_class_kwargs()

        parent_model = kwargs['parent_model']
        del kwargs['parent_model']

        model = kwargs['model']
        del kwargs['model']

        return inlineformset_factory(parent_model, model, **kwargs)

    def get_formset_kwargs(self):
        kwargs = FormsetMixin.get_formset_kwargs(self)
        kwargs['instance'] = self.get_object()
        if 'initial' in kwargs:
            del kwargs['initial']
        if self.use_queryset:
            kwargs['queryset'] = self.get_inline_queryset()
        return kwargs

    def get_inline_queryset(self):
        if self.inline_queryset is None:
            if self.model:
                return self.model._default_manager.all()
            else:
                raise ImproperlyConfigured(u"%(cls)s is missing a inline queryset. Define "
                                           u"%(cls)s.model, %(cls)s.inline_queryset, or override "
                                           u"%(cls)s.get_inline_queryset()." % {
                                                'cls': self.__class__.__name__
                                        })
        return self.inline_queryset._clone()

    def get_success_url(self):
        if self.success_url:
            url = self.success_url % self.object.__dict__
        else:
            try:
                url = self.object.get_absolute_url()
            except AttributeError:
                raise ImproperlyConfigured(
                    "No URL to redirect to.  Either provide a url or define"
                    " a get_absolute_url method on the Model.")
        return url

    def get_context_data(self, **kwargs):
        context = kwargs
        if self.object:
            context['object'] = self.object
            context_object_name = self.get_context_object_name(self.object)
            if context_object_name:
                context[context_object_name] = self.object
        return context


class ProcessFormsetView(View):
    """
    A mixin that processes a formset on POST.
    """
    def get(self, request, *args, **kwargs):
        formset_class = self.get_formset_class()
        formset = self.get_formset(formset_class)
        return self.render_to_response(self.get_context_data(formset=formset))

    def post(self, request, *args, **kwargs):
        formset_class = self.get_formset_class()
        formset = self.get_formset(formset_class)
        if formset.is_valid():
            return self.form_valid(formset)
        else:
            return self.form_invalid(formset)

    # PUT is a valid HTTP verb for creating (with a known URL) or editing an
    # object, note that browsers only support POST for now.
    def put(self, *args, **kwargs):
        return self.post(*args, **kwargs)


class UpdateInlineView(InlineModelFormsetMixin, ProcessFormsetView,
                       TemplateResponseMixin):
    """
    Undate existing inline objects and may be add new
    """
    template_name_suffix = '_formset'


class CreateInlineView(UpdateInlineView):
    """
    Create new inline objects
    """
    use_queryset = True

    def get_inline_queryset(self):
        return UpdateInlineView.get_queryset(self).none()



from setuptools import setup, find_packages


version = __import__('extra_cbv').get_version()


setup(
    name="django-extra-cbv",
    version=version,
    description="Extra class based views for Django",
    keywords="cbv, class based views",
    author="Alex Kamedov",
    author_email="alex@kamedov.ru",
    url="https://github.com/alekam/django-extra-cbv",
    license="New BSD License",
    platforms=["any"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Web Environment",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Utilities"
    ],
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False
)
